(* file: part2.ml
 * author: Bob Muller
 *
 * CS1101 Computer Sciene I
 *
 * Some solutions to part2 of problem set 3, various tessellations.
 *
 * depends: Asai/Uehara Universe Library, 
 *          Cs1101 Library
 *
 * To run, in a Unix shell type:
 *
 * > make
 * > ./part2
 *)
open Image
open Cs1101 (* defines helpers: i2F, i2S, f2I, f2S, randomColor, ... *)
open World
       
(* constants *)

let displayWidth  = 800.                (* window width *)
let displayHeight = 800.                (* window height *)

(* type of world
 *)
type world_t = unit

(* Helpers *******************************************************)

let square side color = Image.rectangle side side color

let empty_scene width height =
  rectangle width height Color.transparent

let emptyImage = empty_scene displayWidth displayHeight

(* Ring ***************************************************** *)

(* ring : float -> float -> Color.t -> Image.t
 *
 * A call (ring radius width color) returns a ring with a radius
 * width and color as given.
 *)
let ring radius width color =
  let outerCircle = (Image.circle radius color) in
  let innerRadius = radius -. width in
  let innerCircle = (Image.circle innerRadius Color.white)
  in
  Image.place_image innerCircle (width, width) outerCircle


(* rowOfRings : int -> Image.t
 * 
 * A call (rowOfRings n) returns a rectangle containing n rings
 * with equal diameters and with random widths and random colors.
 * The height of the rectangle should be the diameter of the rings.  
 *)
let rowOfRings n =
  let diameter = displayWidth /. (i2F n) in
  let radius = diameter /. 2. in
  let emptyBox = empty_scene displayWidth diameter in
  let rec repeat n image =
    match n = 0 with
    | true  -> image
    | false -> let ringWidth = Random.float radius in
	       let color = randomColor() in
	       let x = (i2F (n - 1)) *. diameter in
	       let y = 0.0 in
	       let theRing = ring radius ringWidth color in
	       let newImage = place_image theRing (x, y) image
	       in
	       repeat (n - 1) newImage
  in
  repeat n emptyBox

(* Various Tessellations ********************************************* *)

let ringTessellation n =
  let diameter = displayWidth /. (i2F n) in
  let rec repeat m image =
    match m = 0 with
    | true  -> image
    | false -> let rings = rowOfRings n in
	       let x = 0.0 in
	       let y = (i2F (m - 1)) *. diameter in
	       let newImage = Image.place_image rings (x, y) image 
	       in
	       repeat (m - 1) newImage
  in
  repeat n emptyImage

(* Squares Tessellation ********************************************* *)

let nestedSquare side width color =
  let innerSide = (side -. 2. *. width) in
  let inner = square innerSide Color.white in
  let outer = square side color 
  in
  Image.place_image inner (width, width) outer

let rowOfSquares n =
  let side = displayWidth /. (i2F n) in
  let rec repeat n image =
    match n = 0 with
    | true  -> image
    | false -> let squareWidth = Random.float side in
	       let color = randomColor () in
	       let x = (i2F (n - 1)) *. side in
	       let y = 0.0 in
	       let square = nestedSquare side squareWidth color in
	       let newImage = Image.place_image square (x, y) image
	       in
	       repeat (n - 1) newImage
  in
  repeat n emptyImage

let squareTessellation n =
  let side = displayWidth /. (i2F n) in
  let rec repeat m image =
    match m = 0 with
    | true  -> image
    | false -> let squares = rowOfSquares n in
	       let x = 0.0 in
	       let y = (i2F (m - 1)) *. side in
	       let newImage = Image.place_image squares (x, y) image
	       in
	       repeat (m - 1) newImage
  in
  repeat n emptyImage

(* Hexagons ********************************************************)

let eqlTriangleHeight side =
  sqrt((side ** 2.0) -. (side /. 2.0) ** 2.0)

let hexagon side color =
  let a = side in (* using a, b and c for sides of right triangle. *)
  let b = eqlTriangleHeight side in
  let c = a /. 2. in
  let x2, y2 = (a, 0.) in
  let x3, y3 = (c, b) in
  let x4, y4 = (-.c, b) in
  let x5, y5 = (-.a, 0.) in
  let x6, y6 = (-.c, -.b) in
  let x1, y1 = (c, -.b) in
  let xys = [(x2, y2); (x3, y3); (x4, y4);
	     (x5, y5); (x6, y6); (x1, y1)]
  in
  (polygon xys color)

let rowOfHexagons n =
  let side = displayWidth /. (i2F n) /. 2. in
  let empty = empty_scene displayWidth ((eqlTriangleHeight side) *. 2.) in
  let rec repeat n image =
    match n = 0 with
    | true  -> image
    | false -> let color = randomColor () in
	       let hex = hexagon side color in
	       let x = (i2F (n - 1)) *. (side *. 3.) in
	       let y = 0.0 in
	       let newImage = Image.place_image hex (x, y) image
	       in
	       repeat (n - 1) newImage
  in
  repeat n empty

let hexTessellation n =
  let side = displayWidth /. (i2F n) /. 2. in
  let dy = eqlTriangleHeight side in
  let rec repeat m y image =
    match y > displayHeight with
    | true  -> image
    | false -> let hexagons = rowOfHexagons n in
	       let x = (match m mod 2 = 0 with
			| true  -> -.(side /. 2.)
			| false -> side) in
	       let newImage = Image.place_image hexagons (x, y) image
	       in
	       repeat (m + 1) (y +. dy) newImage
  in
  repeat 0 (-.dy) emptyImage


(* Chickenwire ***************************************************)

let chickenWire side thickness color =
  let innerSide = side -. thickness in
  let hex = hexagon side color in
  place_image (hexagon innerSide Color.blue) (thickness, thickness) hex

let rowOfChickenWire n =
  let side = displayWidth /. (i2F n) /. 2. in
  let empty = empty_scene displayWidth ((eqlTriangleHeight side) *. 2.) in
  let rec repeat n image =
    match n = 0 with
    | true  -> image
    | false -> let color = randomColor () in
	       let thickness = 10.0 in
	       let chicken = chickenWire side thickness color in
	       let x = (i2F (n - 1)) *. (side *. 3.) in
	       let y = 0.0 in
	       let newImage = Image.place_image chicken (x, y) image
	       in
	       repeat (n - 1) newImage
  in
  repeat n empty

let chickenWireTessellation n =
  let side = displayWidth /. (i2F n) /. 2. in
  let dy = eqlTriangleHeight side in
  let rec repeat m y =
    match y > displayHeight with
      true  -> (empty_scene displayWidth displayHeight)
    | false ->
      let x = (match m mod 2 with
		 0 -> -.(side /. 2.)
               | _ -> side)
      in
      place_image (rowOfChickenWire n) (x, y) (repeat (m + 1) (y +. dy))
  in
  repeat 0 (-.dy)

(* ***************************************************************** *)

(* This test oscillates between ring, square and hexagonal tessellations. *)
	 
let variousTessellations world =
  let n = (1 + Random.int 7)
  in
  match (Random.int 4) mod 4 with 
    0 -> ringTessellation n
  | 1 -> squareTessellation n
  | 2 -> hexTessellation n
  | _ -> chickenWireTessellation n

let testTessellation world =
  World.big_bang ()
		 ~name:"Tessellation"
		 ~to_draw:variousTessellations
		 ~width:(f2I displayWidth)
		 ~height:(f2I displayHeight)

let _ = testTessellation ()

